/**
 * http://www.stat.ubc.ca/~bouchard/GEP/README-usage.html
 * 
 * @author Alexandre Bouchard
 * @author Ardavan Saeedi
 */
package gep.comparisons;
import java.util.List;
import java.util.Random;

import nuts.math.GMFct;
import nuts.math.GMFctUtils;
import nuts.math.Graphs;
import nuts.math.RateMtxUtils;
import nuts.math.TabularGMFct;
import nuts.math.TreeSumProd;
import nuts.util.All2OneMap;
import nuts.util.CounterMap;
import nuts.util.MathUtils;
import pty.learn.CTMCExpectations;
import pty.learn.Estimators;
import Jama.Matrix;
import fig.basic.Option;
import gep.GEPMain;
import gep.timeseries.Series;


public class EMForCTMC implements IncrementalReconstructionMethod
{
  @Option public int nSplits = 1;
  
  private List<Series> data;
  private int nChars;
  
  @Override
  public void init(Random samplingRand)
  {
    // sample some random init params
    currentRateParam = new double[nHidden()][nHidden()];
    for (int i = 0; i < nHidden(); i++)
      for (int j = i+1; j < nHidden(); j++)
        currentRateParam[i][j] = currentRateParam[j][i] = samplingRand.nextDouble();
    RateMtxUtils.fillRateMatrixDiagonalEntries(currentRateParam);
    currentStatParam = RateMtxUtils.getStationaryDistribution(currentRateParam);
  }

  @Override
  public void iterate(Random samplingRand, int iterIndex)
  {
    // will hold all the suff stats for the MLE (+1 smoothed)
    Matrix suffStats = Matrix.random(nHidden(), nHidden()).times(1); //new Matrix(nHidden(),nHidden());
    // for each sequence, compute expectations
    // update both suff stats and posteriors for answers
    double logll = 0.0;;
    for (int sIdx = 0; sIdx < data.size(); sIdx++)
    {
      final Series curSeries = data.get(sIdx);
      // construct graphical model
      GMFct<Integer> potentials = createPotentials(data.get(sIdx));
      // compute post
      TreeSumProd<Integer> tsp = new TreeSumProd<Integer>(potentials);
      GMFct<Integer> post = tsp.moments();
      // add the suff stats
      addSuffStat(suffStats, curSeries, post); // will call [][][][] expectations in it
      updateReconstructions(curSeries, post);
      // optionally, update likelihood stats for debug
      logll += tsp.logZ();
    }
    GEPMain.outputManager.printWrite("logll", "Iter", iterIndex, "LogLL", logll);
    
    // re-estimate
    currentRateParam = Estimators.getGeneralRateMatrixMLE(suffStats, 100);
    currentStatParam = RateMtxUtils.getStationaryDistribution(currentRateParam);
    _cachedPr = null;
    _cachedExp = null;
  }

  private void updateReconstructions(Series curSeries, GMFct<Integer> post)
  {
    List<Double> allTimes = curSeries.allTimes();
    curSeries.queryStats = new CounterMap<Integer,Integer>();
    int index = 0;
    for (int i = 0; i < allTimes.size(); i++)
      if (!curSeries.isObserved(allTimes.get(i)))
      {
        for (int hidIdx = 0; hidIdx < nHidden(); hidIdx++)
          curSeries.queryStats.setCount(index, hidIdx, post.get(i, hidIdx));
        index++;
      }
  }

  private GMFct<Integer> createPotentials(Series series)
  {
    List<Double> allTimes = series.allTimes();
    TabularGMFct<Integer> result 
      = GMFctUtils.ones(new TabularGMFct<Integer>(
        Graphs.chainGraph(allTimes.size()), 
        new All2OneMap<Integer,Integer>(nHidden())));
    // obs
    for (int i = 0; i < allTimes.size(); i++)
      if (series.isObserved(allTimes.get(i)))
      {
        int observedIndex = series.observationAt(allTimes.get(i));
        for (int hidIdx = 0; hidIdx < nHidden(); hidIdx++)
          if (hidden2char(hidIdx) != observedIndex)
            result.set(i, hidIdx, 0.0);
      }
    // init distn
    for (int hidIdx = 0; hidIdx < nHidden(); hidIdx++)
      result.set(0, hidIdx, currentStatParam[hidIdx]);
    // trans
    for (int i = 0; i < allTimes.size() - 1; i++)
    {
      final double T = allTimes.get(i+1) - allTimes.get(i);
      double [][] prs = getPr(T);
      for (int hidIdx = 0; hidIdx < nHidden(); hidIdx++)
        for (int hidIdx2 = 0; hidIdx2 < nHidden(); hidIdx2++)
          result.set(i,i+1,hidIdx,hidIdx2, prs[hidIdx][hidIdx2]);
    }
    return result;
  }

  private void addSuffStat(Matrix suffStats, Series curSeries,  GMFct<Integer> post)
  {
    List<Double> allTimes = curSeries.allTimes();
    // for each transition
    for (int i = 0; i < post.graph().vertexSet().size() - 1; i++)
    {
      // check the length
      final double T = allTimes.get(i+1) - allTimes.get(i);
      double [][][][] expss = getExp(T);
      for (int x0 = 0; x0 < nHidden(); x0++)
        for (int xT = 0; xT < nHidden(); xT++)
        {
          double curPost = post.get(i, i+1, x0, xT);
          double [][] currentStat = expss[x0][xT];
          suffStats.plusEquals(new Matrix(currentStat).timesEquals(curPost));
        }
    }
  }

  @Override
  public void loadData(List<Series> data, int nChars)
  {
    this.data = data;
    this.nChars = nChars;
  }

  @Override
  public int reconstruct(int sIdx, int qIdx)
  {
    return data.get(sIdx).queryStats.getCounter(qIdx).argMax();
  }
  
  private double [][] currentRateParam;
  private double [] currentStatParam;
  
  private double [][] _cachedPr = null;
  private double tpr = -1;
  private double [][] getPr(double t)
  {
    if (MathUtils.close(t, tpr) && _cachedPr != null)
      return _cachedPr;
    _cachedPr = RateMtxUtils.marginalTransitionMtx(currentRateParam, t);
    tpr = t;
    return _cachedPr;
  }
  
  private double [][][][] _cachedExp = null;
  private double texp = -1;
  private double [][][][] getExp(double t)
  {
    if (MathUtils.close(t, texp) && _cachedExp != null)
      return _cachedExp;
    _cachedExp = CTMCExpectations.expectations(t, currentRateParam);
    tpr = t;
    return _cachedExp;
  }
  
  private int nHidden() { return nChars * nSplits; }
//  private int hidden2splitIdx(int hidden) { return hidden % nSplits; }
  private int hidden2char(int hidden)  { return hidden / nSplits; }
//  private int charSplit2hidden(int charIdx, int splitIdx)
//  {
//    if (charIdx >= nChars || splitIdx >= nSplits)
//      throw new RuntimeException();
//    return splitIdx + nSplits * charIdx;
//  }
}
