/**
 * http://www.stat.ubc.ca/~bouchard/GEP/README-usage.html
 * 
 * @author Alexandre Bouchard
 * @author Ardavan Saeedi
 */
package gep.data;

import java.io.File;
import java.util.Collections;
import java.util.List;
import java.util.Random;
import java.util.Set;

import ma.MSAParser;
import ma.MSAPoset;
import ma.RateMatrixLoader;
import ma.SequenceType;
import nuts.math.GMFct;
import nuts.util.Arbre;
import nuts.util.CollUtils;
import nuts.util.Indexer;
import pepper.Encodings;
import pty.RootedTree;
import pty.io.Dataset;
import pty.learn.DiscreteBP;
import pty.smc.models.CTMC;
import fig.basic.LogInfo;
import fig.basic.Option;
import gep.timeseries.Measurements;
import goblin.Taxon;


public class RnaDataset implements DataSource
{
  @Option public String alignmentFile;
  @Option public String treeFile;
  @Option public int minNTrans = 2;
//  @Option public Random rand = new Random(1);
  
  private List<Measurements> _data = null;
  private int pointer = 0;
  
  private void ensureLoaded()
  {
    if (_data != null) return;
    _data = CollUtils.list();
    RootedTree tree = RootedTree.Util.load(new File(treeFile));
    MSAPoset msa = MSAParser.parseMSA(new File(alignmentFile));
    Dataset observations = Dataset.DatasetUtils.fromAlignment(new File(alignmentFile), SequenceType.RNA);
    final int nSites = msa.columns().size();
    CTMC ctmc = new CTMC.SimpleCTMC(RateMatrixLoader.k2p(), nSites);
    LogInfo.track("Loading RNA data");
    for (int s = 0; s < nSites; s++)
    {
      LogInfo.logs("Site " + (s+1) + "/" + nSites);
      GMFct<Taxon> post = DiscreteBP.posteriorMarginalTransitions(tree, ctmc, observations, s);
      for (Arbre<Taxon> leaf : tree.topology().leaves())
      {
        Measurements current =createTimeSeries(leaf, post, tree);
        if (nTrans(current) >= minNTrans)
        {
//          System.out.println(current.toString(Encodings.rnaEncodings().nonGapCharactersIndexer()));
          _data.add(current);
        }
      }
    }
    LogInfo.end_track();
//    Collections.shuffle(_data, rand);  BAD! DONE IN GEPRAND INSTEAD
  }

  private int nTrans(Measurements current)
  {
    int result = 0;
    for (int i = 0; i < current.size() - 1; i++)
      if (current.getValue(i) != current.getValue(i+1))
        result++;
    return result;
  }

  private Measurements createTimeSeries(Arbre<Taxon> node, GMFct<Taxon> post, RootedTree tree)
  {
    List<Double> times = CollUtils.list();
    List<Integer> states = CollUtils.list();
    double curTime = 0.0;
    do 
    {
      // find MAP state
      int argmax = -1; double max = Double.NEGATIVE_INFINITY;
      for (int state = 0; state < post.nStates(node.getContents()); state++)
      {
        final double cur = post.get(node.getContents(), state);
        if (cur > max)
        {
          max = cur;
          argmax = state;
        }
      }
      times.add(curTime);
      states.add(argmax);
      //
      curTime += tree.branchLengths().get(node.getContents());
      node = node.getParent();
    } 
    while (!node.isRoot());
    Measurements result = new Measurements(times, states);
//    System.out.println(result.toString(Encodings.rnaEncodings().nonGapCharactersIndexer()));
    return result;
  }

  @Override
  public Set<String> possibleObservations()
  {
    Indexer<Character> charIndex = Encodings.rnaEncodings().nonGapCharactersIndexer();
    Set<String> strIndex = CollUtils.set();
    for (int i = 0; i < charIndex.size(); i++)
      strIndex.add("" + charIndex.i2o(i));
    return strIndex;
  }

  @Override
  public Measurements next(Indexer<String> indexer)
  {
//    if (!indexer.equals(Encodings.rnaEncodings().nonGapCharactersIndexer()))
//      throw new RuntimeException(indexer + " vs " + Encodings.rnaEncodings().nonGapCharactersIndexer());
    ensureLoaded();
    if (pointer++ < _data.size())
      return _data.get(pointer-1);
    else
      return null;
  }
}
