/**
 * http://www.stat.ubc.ca/~bouchard/GEP/README-usage.html
 * 
 * @author Alexandre Bouchard
 * @author Ardavan Saeedi
 */
package gep.model;
import java.util.Random;

import nuts.util.Counter;
import nuts.util.CounterMap;
import fig.basic.Pair;
import gep.util.TP;

/**
 * Compute the predictive distribution of HGEPs
 * 
 * C: ContextType 
 * S: SampledType
 * 
 * Note: in CTMC models, C will be the same as S
 */
public class Predictives<C,S>
{
  public final SufficientStatistics<C,S> readOnlySuffStats;
  public final ContextHierarchy<C> contextHierarchy;
  public final BaseMeasure<C,S> baseMeasure;
  public final HyperParamProvider<C> hyperParams;
  
  //gamma0 will only affect resampling the alphas
  public final double beta0; 
  
  // get the concentration parameter for \bar \mu_context
  private double alpha(C context)
  {
    return hyperParams.getHyperParam(context);
  }
  
  public static Predictives<SplitContext,SplitContext> splitPredictives(      
      SufficientStatistics<SplitContext, SplitContext> readOnlySuffStats,
      double bmAlpha0,
      double alphaLevel1, double alphaLevel2, double alphaLevel3, int nSym, 
      double beta0, double gamma0)
  {
    return new Predictives<SplitContext,SplitContext>(
        readOnlySuffStats, 
        new SplitHierarchy(), 
        new SplitBaseMeasure(bmAlpha0, nSym),
        new SplitHyperParams(alphaLevel1, alphaLevel2, alphaLevel3, nSym, gamma0),
        beta0
        );
  }
  
  public Predictives(
      SufficientStatistics<C, S> readOnlySuffStats,
      ContextHierarchy<C> contextHierarchy, 
      BaseMeasure<C, S> baseMeasure,
      HyperParamProvider<C> hyperParmams, 
      double beta0) 
  {
    this.readOnlySuffStats = readOnlySuffStats;
    this.contextHierarchy = contextHierarchy;
    this.baseMeasure = baseMeasure;
    this.hyperParams = hyperParmams;
    this.beta0 = beta0;
  }

  public Pair<S,Double> sample(Random rand, C context, SufficientStatistics<C,S> readWriteSuffStat)
  { 
    // needs to be called before sampleState()!
    double time = sampleTime(rand, context, readWriteSuffStat);
    S state = sampleState(rand, context, readWriteSuffStat);
    return Pair.makePair(state, time);
  }
  
  @SuppressWarnings("rawtypes")
  private static final Counter DUMMY = new Counter();
  @SuppressWarnings("unchecked")
  public static <C,S> Counter<S> getReadOnly(CounterMap<C,S> map, C key)
  {
    return map.hasCounter(key) ? map.getCounter(key) : DUMMY;
  }
  
  // needs to be called before sampleState()!
  private double sampleTime(
      Random rand, 
      C context, 
      SufficientStatistics<C,S> readWriteSuffStat)
  {
    final Counter<S> currentCacheRW = readWriteSuffStat.getStateTransitionStats().getCounter(context),
                     currentCacheR  = getReadOnly(readOnlySuffStats.getStateTransitionStats(), context); 
    final double alpha = currentCacheRW.totalCount() +
                         currentCacheR.totalCount() +
                         alpha(context);
    final Counter<C> waitingTimeStatsRW = readWriteSuffStat.getWaitingTimeStats(),
                     waitingTimeStatsR  = readOnlySuffStats.getWaitingTimeStats();
    final double beta = waitingTimeStatsRW.getCount(context) +
                        waitingTimeStatsR.getCount(context) + 
                        beta0;
    final double sample = TP.sample(rand, alpha, beta);
    if (!(sample > 0))
      throw new RuntimeException();
    waitingTimeStatsRW.incrementCount(context, sample);
    return sample;
  }

  public S sampleState(
      Random rand, 
      C context, 
      SufficientStatistics<C,S> readWriteSuffStat)
  {
    final Counter<S> currentCacheRW = readWriteSuffStat.getStateTransitionStats().getCounter(context),
                     currentCacheR  = getReadOnly(readOnlySuffStats.getStateTransitionStats(), context); 
    final double nObsInCurContxt = currentCacheRW.totalCount() +
                                   currentCacheR.totalCount();
    final double alpha = alpha(context);
    S result = rand.nextDouble() < nObsInCurContxt / (nObsInCurContxt + alpha) ?
      // use cache
      sampleCache(rand, currentCacheRW, currentCacheR) :
      // sample new
      recurse(rand, context, readWriteSuffStat);
    currentCacheRW.incrementCount(result, 1.0);
    return result;
  }
  
  public static <S> S sampleCache(
      Random rand, 
      Counter<S> currentCacheRW,
      Counter<S> currentCacheR)
  {
    final double totalMass = currentCacheR.totalCount() + currentCacheRW.totalCount();
    final double u = rand.nextDouble() * totalMass;
    double sum = 0.0;
    for (S item : currentCacheRW) { sum += currentCacheRW.getCount(item); if (sum >= u) return item; }
    for (S item : currentCacheR ) { sum += currentCacheR .getCount(item); if (sum >= u) return item; }
    throw new RuntimeException();
  }
  
  private S recurse(
      Random rand, 
      C context,
      SufficientStatistics<C, S> readWriteSuffStat)
  {
    C backoff = contextHierarchy.backoff(context);
    return backoff == null ?
      baseMeasure.sampleState(rand, readWriteSuffStat, readOnlySuffStats) :
      sampleState(rand, backoff, readWriteSuffStat);
  }
}
