/**
 * http://www.stat.ubc.ca/~bouchard/GEP/README-usage.html
 * 
 * @author Alexandre Bouchard
 * @author Ardavan Saeedi
 */
package gep.timeseries;
import gep.GEPMain;
import gep.model.SplitContext;
import gep.model.SufficientStatistics;
import gep.model.SufficientStatisticsImpl;
import gep.pmcmc.Generator;

import java.util.Collections;
import java.util.List;
import java.util.Set;

import nuts.util.CollUtils;
import nuts.util.CounterMap;
import nuts.util.Indexer;
import nuts.util.MathUtils;

/**
 * contains observations and query points, 
 * as well as posterior information for guessing observed state at queries,
 * and sampled hidden states and suff stats for that sample 
 *
 */
public class Series
{
  public final List<Double> queries;
  public final Measurements observations;
  
  public boolean initialized = false;
  public CounterMap<Integer,Integer> queryStats = new CounterMap<Integer,Integer>(); // query index -> char index
  public SufficientStatistics<SplitContext,SplitContext> suffStats 
    = new SufficientStatisticsImpl<SplitContext,SplitContext>(); // initially zero
  public double previousLogLL = Double.NEGATIVE_INFINITY;
  public List<Event<SplitContext>> currentHiddenVariables = null;
  
  public List<Double> allTimes()
  {
    Set<Double> answer = CollUtils.set();
    answer.addAll(queries);
    answer.addAll(observations.getTimes());
    if (answer.size() != queries.size() + observations.size())
      throw new RuntimeException();
    List<Double> ordered = CollUtils.list(answer);
    Collections.sort(ordered);
    return ordered;
  }
  
  
  
  public Series(Measurements observations, List<Double> queries)
  {
    this.queries = queries;
    this.observations = observations;
  }
  public CounterMap<Integer,Integer> getQueryPosteriors()
  {
    CounterMap<Integer,Integer> result = new CounterMap<Integer,Integer>();
    result.incrementAll(queryStats);
    result.normalize();
    return result;
  }
  public void updateQueryStats()
  {
//    final double init = queryStats.totalCount();
    for (int queryIndex = 0; queryIndex < queries.size(); queryIndex++)
    {
      boolean found = false;
      final double queryPosition = queries.get(queryIndex); 
      loop:for (int hidIndex = 1; hidIndex < currentHiddenVariables.size(); hidIndex++)
      {
        Event<SplitContext> e1 = currentHiddenVariables.get(hidIndex-1),
                            e2 = currentHiddenVariables.get(hidIndex);
        if (Event.left(e1, e2) <= queryPosition && queryPosition < Event.right(e1,e2))
        {
          final int curChar = Event.contents(e1, e2).getX();
          queryStats.incrementCount(queryIndex, curChar, 1.0);
          found = true;
          break loop;
        }
      }
      if (!found)
        throw new RuntimeException("None matched queryIndex=" + queryIndex + ", pos=" + queryIndex + "\n"
            + "Series:\n" + Generator.toString(currentHiddenVariables, new Indexer<Double>(), GEPMain.observedIndexer));
    }
  }

  public boolean isObserved(double d)
  {
    if (_isObserved(d)) return true;
    if (!_isQuery(d)) throw new RuntimeException();
    return false;
  }
  
  private boolean _isObserved(double d)
  {
    for (double d2 : observations.getTimes())
      if (MathUtils.close(d, d2))
        return true;
    return false;
  }
  private boolean _isQuery(double d)
  {
    for (double d2 : queries)
      if (MathUtils.close(d, d2))
        return true;
    return false;
  }



  public int observationAt(double double1)
  {
    // find the associated index
    for (int i = 0; i < observations.size(); i++)
      if (MathUtils.close(double1, observations.getTime(i)))
        return observations.getValue(i);
    throw new RuntimeException();
  }
}